#Příliš žluťoučký kůň úpěl ďábelské ó - PŘÍLIŠ ŽLUŤOUČKÝ KŮŇ ÚPĚL ĎÁBELSKÉ Ó
#P:/60_PGM/_SRC/m26_00_DescPropSlot.py
"""
Dokumentační komentář modulu.
"""



#Výpis 26.1: Definice třídy descriptor
############################################################################
# >>> \
class descriptor:
    """Datový deskriptor průběžně zaznamenávající jednotlivé přístupy."""

    def __init__(self, initval=None):
        print(f'Vytvořen deskriptor s výchozí hodnotou «{initval}»')
        self.pool = {None: initval}

    def __set_name__(self, owner, name):
        self.name = name
        print('Přiřazuje se název:', name, '-  Vlastník:', owner)

    def __get__(self, instance, objtype):
        val = self                if instance is None else \
              self.pool[instance] if instance in self.pool.keys() else \
              self.pool[None]
        print(f"Čtení: atribut {self.name} objektu {instance} "
              f"typu {objtype} = {val}")
        return val

    def __set__(self, instance, val):
        print(f"Zápis atributu {self.name} objektu {instance} = {val}")
        self.pool[instance] = val

    def __delete__(self, instance):
        print(f"Odstranění atributu {self.name} objektu {instance}")
        self.pool.pop(instance)



#Výpis 26.2: Hrátky s deskriptorem
############################################################################
# >>> \
class C25b:
    a1 = descriptor('První')
    a2 = descriptor('Druhý')
    def __init__(self, id):   self.id = id
    def __repr__(self):   return 'C25b_' + str(self.id)
    print('Třída C25b definována')

# Vytvořen deskriptor s výchozí hodnotou «První»
# Vytvořen deskriptor s výchozí hodnotou «Druhý»
# Třída C25b definována
# Přiřazuje se název: a1 -  Vlastník: <class '__main__.C25b'>
# Přiřazuje se název: a2 -  Vlastník: <class '__main__.C25b'>
c1 = C25b(1);   c2 = C25b(2);   C25b.a1,   c1.a1,   c2.a1
# Čtení: atribut a1 objektu None typu <class '__main__.C25b'> = <__main__.descriptor object at 0x000001C4EFB95DF0>
# Čtení: atribut a1 objektu C25b_1 typu <class '__main__.C25b'> = První
# Čtení: atribut a1 objektu C25b_2 typu <class '__main__.C25b'> = První
# (<__main__.descriptor object at 0x000001C4EFB95DF0>, 'První', 'První')
c1.a1='C1';   c2.a1='C2';   C25b.a1, c1.a1, c2.a1, c1.a2, c2.a2
# Zápis atributu a1 objektu C25b_1 = C1
# Zápis atributu a1 objektu C25b_2 = C2
# Čtení: atribut a1 objektu C25b_1 typu <class '__main__.C25b'> = C1
# Čtení: atribut a1 objektu C25b_2 typu <class '__main__.C25b'> = C2
# Čtení: atribut a2 objektu C25b_1 typu <class '__main__.C25b'> = Druhý
# Čtení: atribut a2 objektu C25b_2 typu <class '__main__.C25b'> = Druhý
# ('První', 'C1', 'C2', 'Druhý', 'Druhý')
C25b.a1='Třídní';   C25b.a1, c1.a1, c2.a1, c1.a2, c2.a2
# Čtení: atribut a2 objektu C25b_1 typu <class '__main__.C25b'> = Druhý
# Čtení: atribut a2 objektu C25b_2 typu <class '__main__.C25b'> = Druhý
# ('Třídní', 'Třídní', 'Třídní', 'Druhý', 'Druhý')
# >>>



#Výpis 26.3: Deskriptor vytvářený vnitřní třídou
############################################################################
# >>> \
class C25c:
    def __init__(self, val):
        self._atr = val

    class descr:
        def __get__(self, instance, owner):
            return f"Přepočtená hodnota «{instance._atr}»"
        def __set__(self, instance, value):
            raise AttributeError('Atribut atr nelze měnit')
        def __delete__(self, instance, value):
            raise AttributeError('Atribut atr nelze odstranit')

    atr = descr()

cc1=C25c('První');   cc2=C25c('Druhý')
cc1.atr,   cc2.atr
# ('Přepočtená hodnota «První»', 'Přepočtená hodnota «Druhý»')
cc1.atr = 0
# Traceback (most recent call last):
#   File "<pyshell#16>", line 1, in <module>
#     cc1.atr = 0
#   File "<pyshell#12>", line 10, in __set__
#     raise AttributeError('Atribut atr nelze měnit')
# AttributeError: Atribut atr nelze měnit
# >>>



#Výpis 26.4: Deskriptor pro odložené vyhodnocení
############################################################################
# >>> \
class reify(object):
    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __get__(self, inst, objtype=None):
        if inst is None:
            return self
        val = self.wrapped(inst)
        setattr(inst, self.wrapped.__name__, val)
        return val

# >>> \
class C26c:
    @reify
    def attr(self):
        print('Počítám hodnotu atributu...')
        return 'Spočteno'

cc = C26c();   cc.attr
# Počítám hodnotu atributu...
# 'Spočteno'
cc.attr, cc.attr
# ('Spočteno', 'Spočteno')
# >>>



#Výpis 26.5 Vlastnost jako instance třídy property
############################################################################
# >>> \
class C26a:
    def __init__(self, id, val):   self._id = id;   self._attr = val
    def __repr__(self):      return f"C26a({self._id})"
    def getAttr (self):      print(f"Čtení {self}:"); return self._attr
    def setAttr (self, val): print(f"Zápis {self}:"); self._attr = val
    # def delAttr (self):      print(f"Mazat {self}:"); del self._attr
    # attr = property(getAttr, setAttr, delAttr, 'Demonstrační atribut')
    attr = property(getAttr, setAttr, doc='Demonstrační atribut')

c1 = C26a(1, 'Raz');   c2 = C26a(2, 'Dva');   c1, c2
# (C26a(1), C26a(2))
c1.attr='One';   c2.attr='Two';   c1.attr, c2.attr
# Zápis C26a(1):
# Zápis C26a(2):
# Čtení C26a(1):
# Čtení C26a(2):
# ('One', 'Two')
help(C26a.attr)
# Help on property:

#     Demonstrační atribut

# >>>



#Výpis 26.6: Ekvivalent definice zabudované třídy property
############################################################################

""" Je uloženo v souboru m26_02_Property.py. """



#Výpis 26.7: Zadání vlastnosti pomocí dekorátoru
############################################################################
# >>> \
class C26b:
    def __init__(self, id, val):
        self._id = id;
        self._attr = val

    def __repr__(self):
        return f"C26a({self._id}-{self._attr})"

    @property
    def attr(self):
        """Dokumentační komentář vlastnosti C26b.attr."""
        print(f"Čtení {self}.attr (= {self._attr})");
        return self._attr

    @attr.setter
    def attr(self, val):
        print(f"Zápis {self}.attr ({self._attr}->{val})")
        self._attr = val

    @attr.deleter
    def attr(self):
        print(f"Mazat {self}.attr (= {self._attr})")
        del self._attr

c1 = C26b(1, 'Raz');   c2 = C26b(2, 'Dva');   c1, c2
# ((C26a(1-Raz), C26a(2-Dva))
c1.attr='One';   c2.attr='Two';   c1.attr, c2.attr
# Zápis C26a(1-Raz).attr (Raz->One)
# Zápis C26a(2-Dva).attr (Dva->Two)
# Čtení C26a(1-One).attr (= One)
# Čtení C26a(2-Two).attr (= Two)
# ('One', 'Two')
help(C26b.attr)
# Help on property:

    # Dokumentační komentář vlastnosti C26b.attr.

# >>>



#Výpis 26.8: Ukázka používání slotů
############################################################################
# >>> \
class Point:
    __slots__ = ("x", "y")

    def __init__(self, x=0, y=0):
        self.x = x;
        self.y = y

    def __repr__(self):
        return f"Point(x={self.x}, y={self.y})"

p = Point(1, 2);   p
# Point(x=1, y=2)
p.x = 3;   p
# Point(x=3, y=2)
p.z = 4
# Traceback (most recent call last):
#   File "<pyshell#74>", line 1, in <module>
#     p.z = 4
# AttributeError: 'Point' object has no attribute 'z'
# >>>



############################################################################
##### KONEC #####
